#include <iostream>

#include "serializationmanager.h"
#include "serializable.h"

#include <sstream>
#include <fstream>

/**
 * @brief Plop = Serializable mother class (incomplete)
 */
class Plop : public Serializable
{
public:
    P_REFERENCE   (Plop, myPlop)
};

/**
 * @brief Bar = Serializable class Inheriting Plop
 */
class Bar : public Plop
{
public:
    P_STRING      (myTruc)

    SERIALIZABLE(Bar,
                 CAST(myTruc),
                 CAST(myPlop))

    Bar()
    {
        myTruc = "Je suis un Bar";
        myPlop = this;
    }
};

/**
 * @brief Bar = Other Serializable class Inheriting Plop, referencing an instance of the Bar class
 */
class Foo : public Plop
{
public:
    P_INT         (myInt)
    P_STRING      (myString)
    P_VECTOR_FLOAT(myFloatArray)
    P_VEC3        (myVec3)

    SERIALIZABLE(Foo,
                 CAST(myInt),
                 CAST(myString),
                 CAST(myFloatArray),
                 CAST(myVec3),
                 CAST(myPlop))

    Foo()
    {
        myInt = 3;
        myString = "Je suis un Foo";
        myFloatArray.push_back(1.5);
        myFloatArray.push_back(1.3);
        myFloatArray.push_back(2.5);
        myFloatArray.push_back(6.2);
        myVec3 = glm::vec3(0.1, 0, 5.4);
        Bar* bar = new Bar();
        bar->myPlop = nullptr;
        myPlop = bar;
    }
};

INIT_SERIALIZABLE(Bar)
INIT_SERIALIZABLE(Foo)

int main()
{
    // testing polymorphism on Serializables
    Foo* foo = new Foo();

    // print before
    std::cout << "Before :" << std::endl;
    ObjectSaver printer;
    printer.addObject(foo);
    printer.saveAscii(std::cout);

    // saving as binary
    ObjectSaver saverBinary;
    saverBinary.addObject(foo);
    std::fstream outFileBinary;
    outFileBinary.open("test.bin", std::ios_base::out);
    saverBinary.saveBinary(outFileBinary);
    outFileBinary.close();

    // loading as binary
    ObjectLoader loaderBinary;
    std::fstream inFileBinary;
    inFileBinary.open("test.bin", std::ios_base::in);
    loaderBinary.loadBinary(inFileBinary);
    inFileBinary.close();

    // saving as json
    ObjectSaver saverAscii;
    for(Foo* f : loaderBinary.getObjects<Foo>())
        saverAscii.addObject(f);
    std::fstream outFileAscii;
    outFileAscii.open("test.txt", std::ios_base::out);
    saverAscii.saveAscii(outFileAscii);
    outFileAscii.close();

    // loading as json
    ObjectLoader loaderAscii;
    std::fstream inFileAscii;
    inFileAscii.open("test.txt", std::ios_base::in);
    loaderAscii.loadAscii(inFileAscii);
    inFileAscii.close();

    // print after
    std::cout << "After :" << std::endl;
    printer.clearObjects();
    for(Foo* f : loaderAscii.getObjects<Foo>())
        printer.addObject(f);
    printer.saveAscii(std::cout);

    return 0;
}
