#include "property.h"
#include "serializable.h"

// INTEGER PROPERTY
template <> std::ostream& Property<int>::saveBinary(std::ostream& os) { return os.write((char*)&m_value, sizeof(int)); }
template <> std::ostream& Property<int>::saveAscii(std::ostream& os) { return os << m_value; }
template <> std::istream& Property<int>::loadBinary(std::istream& is) { return is.read((char*)&m_value, sizeof(int)); }
template <> std::istream& Property<int>::loadAscii(std::istream& is) { return is >> m_value; }
template <> AbstractProperty::PropertyType Property<int>::getStaticPropertyType() { return AbstractProperty::INTEGER; }

// FLOAT PROPERTY
template <> std::ostream& Property<float>::saveBinary(std::ostream& os) { return os.write((char*)&m_value, sizeof(float)); }
template <> std::ostream& Property<float>::saveAscii(std::ostream& os) { return os << m_value; }
template <> std::istream& Property<float>::loadBinary(std::istream& is) { return is.read((char*)&m_value, sizeof(float)); }
template <> std::istream& Property<float>::loadAscii(std::istream& is) { return is >> m_value; }
template <> AbstractProperty::PropertyType Property<float>::getStaticPropertyType() { return AbstractProperty::FLOAT; }

// BOOLEAN PROPERTY
template <> std::ostream& Property<bool>::saveBinary(std::ostream& os) { return os.write((char*)&m_value, sizeof(bool)); }
template <> std::ostream& Property<bool>::saveAscii(std::ostream& os) { return os << m_value; }
template <> std::istream& Property<bool>::loadBinary(std::istream& is) { return is.read((char*)&m_value, sizeof(bool)); }
template <> std::istream& Property<bool>::loadAscii(std::istream& is) { return is >> m_value; }
template <> AbstractProperty::PropertyType Property<bool>::getStaticPropertyType() { return AbstractProperty::BOOLEAN; }

// UNSIGNED INTEGER PROPERTY
template <> std::ostream& Property<unsigned int>::saveBinary(std::ostream& os) { return os.write((char*)&m_value, sizeof(unsigned int)); }
template <> std::ostream& Property<unsigned int>::saveAscii(std::ostream& os) { return os << m_value; }
template <> std::istream& Property<unsigned int>::loadBinary(std::istream& is) { return is.read((char*)&m_value, sizeof(unsigned int)); }
template <> std::istream& Property<unsigned int>::loadAscii(std::istream& is) { return is >> m_value; }
template <> AbstractProperty::PropertyType Property<unsigned int>::getStaticPropertyType() { return AbstractProperty::UNSIGNED_INTEGER; }

// BYTE PROPERTY
template <> std::ostream& Property<char>::saveBinary(std::ostream& os) { return os.write((char*)&m_value, sizeof(char)); }
template <> std::ostream& Property<char>::saveAscii(std::ostream& os) { return os << m_value; }
template <> std::istream& Property<char>::loadBinary(std::istream& is) { return is.read((char*)&m_value, sizeof(char)); }
template <> std::istream& Property<char>::loadAscii(std::istream& is) { return is >> m_value; }
template <> AbstractProperty::PropertyType Property<char>::getStaticPropertyType() { return AbstractProperty::BYTE; }

// SHORT PROPERTY
template <> std::ostream& Property<short>::saveBinary(std::ostream& os) { return os.write((char*)&m_value, sizeof(short)); }
template <> std::ostream& Property<short>::saveAscii(std::ostream& os) { return os << m_value; }
template <> std::istream& Property<short>::loadBinary(std::istream& is) { return is.read((char*)&m_value, sizeof(short)); }
template <> std::istream& Property<short>::loadAscii(std::istream& is) { return is >> m_value; }
template <> AbstractProperty::PropertyType Property<short>::getStaticPropertyType() { return AbstractProperty::SHORT; }

// LONG PROPERTY
template <> std::ostream& Property<long>::saveBinary(std::ostream& os) { return os.write((char*)&m_value, sizeof(long)); }
template <> std::ostream& Property<long>::saveAscii(std::ostream& os) { return os << m_value; }
template <> std::istream& Property<long>::loadBinary(std::istream& is) { return is.read((char*)&m_value, sizeof(long)); }
template <> std::istream& Property<long>::loadAscii(std::istream& is) { return is >> m_value; }
template <> AbstractProperty::PropertyType Property<long>::getStaticPropertyType() { return AbstractProperty::LONG; }

// DOUBLE PROPERTY
template <> std::ostream& Property<double>::saveBinary(std::ostream& os) { return os.write((char*)&m_value, sizeof(double)); }
template <> std::ostream& Property<double>::saveAscii(std::ostream& os) { return os << m_value; }
template <> std::istream& Property<double>::loadBinary(std::istream& is) { return is.read((char*)&m_value, sizeof(double)); }
template <> std::istream& Property<double>::loadAscii(std::istream& is) { return is >> m_value; }
template <> AbstractProperty::PropertyType Property<double>::getStaticPropertyType() { return AbstractProperty::DOUBLE; }

// VEC2 PROPERTY
template <> std::ostream& Property<glm::vec2>::saveBinary(std::ostream& os) { return os.write((char*)glm::value_ptr(m_value), sizeof(float)*m_value.length()); }
template <> std::ostream& Property<glm::vec2>::saveAscii(std::ostream& os) { os << "["; for(int i=0; i<m_value.length(); ++i) os << (glm::value_ptr(m_value)[i]) << (i == m_value.length()-1 ? "]" : ", "); return os; }
template <> std::istream& Property<glm::vec2>::loadBinary(std::istream& is) { return is.read((char*)glm::value_ptr(m_value), sizeof(float)*m_value.length()); }
template <> std::istream& Property<glm::vec2>::loadAscii(std::istream& is) { is.ignore(1); for(int i=0; i<m_value.length(); ++i) is >> (glm::value_ptr(m_value)[i]), is.ignore(1); return is; }
template <> AbstractProperty::PropertyType Property<glm::vec2>::getStaticPropertyType() { return AbstractProperty::VEC2; }

// VEC3 PROPERTY
template <> std::ostream& Property<glm::vec3>::saveBinary(std::ostream& os) { return os.write((char*)glm::value_ptr(m_value), sizeof(float)*m_value.length()); }
template <> std::ostream& Property<glm::vec3>::saveAscii(std::ostream& os) { os << "["; for(int i=0; i<m_value.length(); ++i) os << (glm::value_ptr(m_value)[i]) << (i == m_value.length()-1 ? "]" : ", "); return os; }
template <> std::istream& Property<glm::vec3>::loadBinary(std::istream& is) { return is.read((char*)glm::value_ptr(m_value), sizeof(float)*m_value.length()); }
template <> std::istream& Property<glm::vec3>::loadAscii(std::istream& is) { is.ignore(1); for(int i=0; i<m_value.length(); ++i) is >> (glm::value_ptr(m_value)[i]), is.ignore(1); return is; }
template <> AbstractProperty::PropertyType Property<glm::vec3>::getStaticPropertyType() { return AbstractProperty::VEC3; }

// VEC4 PROPERTY
template <> std::ostream& Property<glm::vec4>::saveBinary(std::ostream& os) { return os.write((char*)glm::value_ptr(m_value), sizeof(float)*m_value.length()); }
template <> std::ostream& Property<glm::vec4>::saveAscii(std::ostream& os) { os << "["; for(int i=0; i<m_value.length(); ++i) os << (glm::value_ptr(m_value)[i]) << (i == m_value.length()-1 ? "]" : ", "); return os; }
template <> std::istream& Property<glm::vec4>::loadBinary(std::istream& is) { return is.read((char*)glm::value_ptr(m_value), sizeof(float)*m_value.length()); }
template <> std::istream& Property<glm::vec4>::loadAscii(std::istream& is) { is.ignore(1); for(int i=0; i<m_value.length(); ++i) is >> (glm::value_ptr(m_value)[i]), is.ignore(1); return is; }
template <> AbstractProperty::PropertyType Property<glm::vec4>::getStaticPropertyType() { return AbstractProperty::VEC4; }

// MAT3 PROPERTY
template <> std::ostream& Property<glm::mat3>::saveBinary(std::ostream& os) { return os.write((char*)glm::value_ptr(m_value), sizeof(float)*m_value.length()); }
template <> std::ostream& Property<glm::mat3>::saveAscii(std::ostream& os) { os << "["; for(int i=0; i<m_value.length(); ++i) os << (glm::value_ptr(m_value)[i]) << (i == m_value.length()-1 ? "]" : ", "); return os; }
template <> std::istream& Property<glm::mat3>::loadBinary(std::istream& is) { return is.read((char*)glm::value_ptr(m_value), sizeof(float)*m_value.length()); }
template <> std::istream& Property<glm::mat3>::loadAscii(std::istream& is) { is.ignore(1); for(int i=0; i<m_value.length(); ++i) is >> (glm::value_ptr(m_value)[i]), is.ignore(1); return is; }
template <> AbstractProperty::PropertyType Property<glm::mat3>::getStaticPropertyType() { return AbstractProperty::MAT3; }

// MAT4 PROPERTY
template <> std::ostream& Property<glm::mat4>::saveBinary(std::ostream& os) { return os.write((char*)glm::value_ptr(m_value), sizeof(float)*m_value.length()); }
template <> std::ostream& Property<glm::mat4>::saveAscii(std::ostream& os) { os << "["; for(int i=0; i<m_value.length(); ++i) os << (glm::value_ptr(m_value)[i]) << (i == m_value.length()-1 ? "]" : ", "); return os; }
template <> std::istream& Property<glm::mat4>::loadBinary(std::istream& is) { return is.read((char*)glm::value_ptr(m_value), sizeof(float)*m_value.length()); }
template <> std::istream& Property<glm::mat4>::loadAscii(std::istream& is) { is.ignore(1); for(int i=0; i<m_value.length(); ++i) is >> (glm::value_ptr(m_value)[i]), is.ignore(1); return is; }
template <> AbstractProperty::PropertyType Property<glm::mat4>::getStaticPropertyType() { return AbstractProperty::MAT4; }

// STRING PROPERTY
template <> std::ostream& Property<std::string>::saveBinary(std::ostream& os)
{
    static const char endChar = '\0';
    os.write(m_value.data(), sizeof(char)*m_value.size());
    return os.write(&endChar, sizeof(char));;
}
template <> std::ostream& Property<std::string>::saveAscii(std::ostream& os)
{
    os << "\"";
    for(char c : m_value)
    {
        switch(c)
        {
        case '\t' :
            os << "\\t";
            break;
        case '\\' :
            os << "\\\\";
            break;
        case '\n' :
            os << "\\n";
            break;
        case '\r' :
            os << "\\r";
            break;
        case '"' :
            os << "\\\"";
            break;
        default:
            os << c;
        }
    }
    return os << "\"";
}
template <> std::istream& Property<std::string>::loadBinary(std::istream& is)
{
    m_value = "";
    for(char c = is.get(); c != '\0'; c = is.get())
        m_value.push_back(c);
    return is;
}
template <> std::istream& Property<std::string>::loadAscii(std::istream& is)
{
    m_value = "";
    is.ignore(1); // "
    for(char c = is.get(); c != '"'; c = is.get())
    {
        if(c == '\\')
        {
            c = is.get();
            switch(c)
            {
            case 't' :
                m_value.push_back('\t');
                break;
            case '\\' :
                m_value.push_back('\\');
                break;
            case 'n' :
                m_value.push_back('\n');
                break;
            case 'r' :
                m_value.push_back('\r');
                break;
            case '"' :
                m_value.push_back('"');
                c = 'c';
                break;
            default:
                m_value.push_back(c);
            }
        }
        else
            m_value.push_back(c);
    }
    return is;
}
template <> AbstractProperty::PropertyType Property<std::string>::getStaticPropertyType() { return AbstractProperty::STRING; }

// ARRAY PROPERTY SPECIAL BOOL HANDLING
template <> std::ostream& ArrayProperty<bool>::saveBinary(std::ostream& os)
{
    unsigned int size = m_value.size();
    os.write((char*)&size, sizeof(unsigned int));
    for(bool it : m_value)
        Property<bool>(it).saveBinary(os);
    return os;
}

template <> std::ostream& ArrayProperty<bool>::saveAscii(std::ostream& os)
{
    os << "{ " << m_value.size() << ", [" << std::endl;
    for(bool it : m_value)
    {
        Property<bool>(it).saveAscii(os);
        os << "," << std::endl;
    }
    return os << "] }";
}
