#ifndef TASKMANAGER_H
#define TASKMANAGER_H

#include <string>
#include <list>
#include <vector>
#include <thread>
#include <mutex>

class Task
{
public:
    enum TaskTypeFlags
    {
        HDD, // saving of loading to the file system
        CPU, // computing something
        GPU, // computing something on the GPU or transferring data between the RAM and the GPU RAM
        NET  // sending or receiving from the network
    };

    virtual void performTask() = 0;

    virtual int getProgressInt() = 0;
    virtual int getProgressTotal() { return 1; }
    virtual float getProgressProportion() { return float(getProgressInt())/float(getProgressTotal()); }

    virtual TaskTypeFlags getFlags() = 0;
    virtual const std::string &getLabel() = 0;
};

class TaskThread
{
    std::list<Task*> m_tasks;
    std::vector<std::thread*> m_threads;
    std::vector<Task*> m_currentTasks;
    std::mutex m_mutex;
    int m_tasksPerformed;
    int m_totalTasks;

    void run(int id);
public:
    ~TaskThread();

    void addTask(Task* task);
    void start(int nbThreads = 1);
    void waitUntilOver();
    bool isOver();
    float getProgressProportion();
};

class TaskManager
{
    TaskThread m_hddTasks;
    TaskThread m_cpuTasks;
    TaskThread m_gpuTasks;
    TaskThread m_netTasks;

public:
    void addTask(Task* task);
    void start();
    void waitUntilOver();
    bool isOver();
    float getProgress(int flags = Task::CPU | Task::GPU | Task::HDD | Task::NET);
};

#endif // TASKMANAGER_H
